#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "rand.h"

#define MAX_AN  10
#define MAX_MN  11

#define DIST_BIN 100

void usage()
{
	printf("\nUSAGE:\n mufu {-N NUM | -M MASS} [-hr] -m M1 [-a ALPHA1 -m M2 [..]]\n\n"
		"OPTIONS:\n"
		" -N  specify number of stars\n"
		" -M  specify total mass of the cluster (in M_sun)\n"
		" -h  print this help and exit\n"
		" -r  use random scheme for stellar masses; default is 'deterministic' scheme\n"
		" -m  specify limiting masses of the mass function; this option may be used\n"
		"     repeatedly, each value determining a break in the power-law mass function.\n"
		" -a  specify index of the mass function between two limiting masses\n\n"
		"EXAMPLES:\n"
		" - generate 1000 masses according to the Salpeter's mass function in <0.1,100>:\n\n"
		"   mufu -N 1000 -r -m 0.1 -a -2.35 -m 100\n\n"
		" - generate Kroupa's favourite mass function:\n\n"
		"   mufu -N 1000 -r -m 0.075 -a -1.3 -m 0.5 -a -2.3 -m 150\n\n"
		" - generate single mass set:\n\n"
		"   mufu -N 1000 -m 1\n\n");
};

double subint(double min, double max, double alpha)
{
//	alpha++;
	if (alpha == 0.) return log(max / min);
	else return (pow(max, alpha) - pow(min, alpha)) / alpha;
};

double mlow(double mhigh, double alpha, double norma, double delta)
{
	if (alpha == 0.) return mhigh / exp(delta / norma);
	else return pow(pow(mhigh, alpha) - delta * alpha / norma, 1. / alpha);
};

int main(int argc, char **argv)
{
	long   i, j, option;
	int    an, mn;
	int    N_tot;
	int    random;
//	int    distrib[DIST_BIN];
	double a, n, tmp, mup, ml;
	double mass, single_mass;
	double alpha[MAX_AN], mlim[MAX_MN];
	double submass[MAX_AN], subcount[MAX_AN], norm[MAX_AN];
	double M_tot, M_tmp, N_tmp;
	
	single_mass = -1.;
	M_tot = N_tot = 0;
	an = mn = 0;
	random = 0;
//	for (i = 0; i < DIST_BIN; i++) distrib[i] = 0;
	
	while ((option = getopt(argc, argv, "N:M:a:m:rh")) != -1) switch (option)
	 {
		case 'M' :
				M_tot = atof(optarg); break;
		case 'N' :
				N_tot = atol(optarg); break;
		case 'a' :
				if (an < MAX_AN) { alpha[an] = atof(optarg); an++; break; }
				else { fprintf(stderr, "Error: Number of alphas exceded maximum limit of %d\n", MAX_AN); return 1; };
		case 'm' :
				if (mn < MAX_MN)
				 {
					mlim[mn] = atof(optarg);
					if (mn && mlim[mn] <= mlim[mn-1])
					 {
						fprintf(stderr, "Error: limiting masses must be given in an ascending order");
					 };
					mn++; break;
				 }
				else { fprintf(stderr, "Error: Number of mass params exceded maximum limit of %d\n", MAX_MN); return 1; };
		case 'h' :
				usage(); return 0;
		case 'r' :
				random = 1; break;
		case ':' :
		case '?' :
			usage();
			return 1;
	 };
	
	if (an >= mn) an = mn - 1;
	mn = an + 1;
	if (!mn)
	 {
		fprintf(stderr, "Error: at least one mass limit has to be specified\n");
		return 1;
	 }
	else if (mn == 1)
	 {
		single_mass = mlim[0];
		N_tmp = 1;
		M_tmp = single_mass;
	 }
	else
	 {
		norm[an-1] = 1.;
		N_tmp = subcount[an-1] = subint(mlim[an-1], mlim[an], alpha[an-1] + 1.);
		M_tmp = submass[an-1] = subint(mlim[an-1], mlim[an], alpha[an-1] + 2.);
		for (i = an - 2; i >= 0; i--)
		 {
			norm[i] = norm[i+1] * pow(mlim[i+1], alpha[i+1] - alpha[i]);
			subcount[i] = norm[i] * subint(mlim[i], mlim[i+1], alpha[i] + 1.);
			N_tmp += subcount[i];
			submass[i] = norm[i] * subint(mlim[i], mlim[i+1], alpha[i] + 2.);
			M_tmp += submass[i];
		 };
	 };
	 
	if ((N_tot == 0) && (M_tot == 0.))
	 {
		fprintf(stderr, "Error: either total mass or number of stars has to be specified\n");
		return 1;
	 }
	else if (N_tot > 0)
	 {
		M_tot = (M_tmp * N_tot) / N_tmp;
	 }
	else if (M_tot > 0.)
	 {
		N_tot = (int)(N_tmp * M_tot / M_tmp);
		M_tot = (M_tmp * N_tot) / N_tmp;
	 }
	else
	 {
//		single_mass = M_tot / N_tot;
		fprintf(stderr, "Error: options -M and -N are mutually exclusive\n");
		return 1;
	 };
	
	if (single_mass > 0.)
	 {
		for (i = 0; i < N_tot; i++) printf("%8.4f\n", single_mass);
	 }
	else
	 {
		for (i = 0; i < an; i++)
		 {
			printf("# <%.2f , %.2f> .. %.2f\n", mlim[i], mlim[i+1], alpha[i]);
			norm[i] *= M_tot / M_tmp;
		 };
	
		if (random)
		 {
			time(&i); // inicializace generatoru nahodnych cisel
			init_rand(i);
			rand2();
			for (i = 1; i < an; i++) subcount[i] += subcount[i-1];
			for (i = 0; i < N_tot; i++)
			 {
				tmp = rand2() * INV_RAND_MAX2 * subcount[an-1];
				for (j = 0; (j < an) && (subcount[j] < tmp); j++);
				if (alpha[j] != -1.)
				 {
					ml = pow(mlim[j], 1. + alpha[j]);
					mup = pow(mlim[j+1], 1. + alpha[j]);
				 }
				else
				 {
					ml = log(mlim[j]);
					mup = log(mlim[j+1]);
				 };
				tmp = ml + rand2() * INV_RAND_MAX2 * (mup - ml);
				if (alpha[j] != -1.) mass = pow(tmp, 1. / (1. + alpha[j]));
				else mass = exp(tmp);
				printf("%8.4f\n", mass);
/*				j = (int)(log(mass / mlim[0]) / log(mlim[an] / mlim[0]) * DIST_BIN);
				distrib[j]++; */
			 };
		 }
		else
		 {
			M_tmp = 0.;
			mup = mlim[an];
			a = alpha[an-1];
			n = norm[an-1];
			for (i = 0; (i < N_tot) && (an > 0); i++)
			 {
				ml = mlow(mup, a + 1., n, 1.);
				if (ml < mlim[an-1])
				 {
					mass = n * subint(mlim[an-1], mup, a + 2.);
					tmp = 1. - n * subint(mlim[an-1], mup, a + 1.);
					an--;
					mup = mlim[an];
					if (an > 0)
					 {
						a = alpha[an-1];
						n = norm[an-1];
						ml = mlow(mup, a + 1., n, tmp);
					 };
				 }
				else mass = 0.;
				if (an) mass += n * subint(ml, mup, a + 2.);
				printf("%8.4f\n", mass);
/*				j = (int)(log(mass / mlim[0]) / log(mlim[mn-1] / mlim[0]) * DIST_BIN);
				if (j >= DIST_BIN) distrib[DIST_BIN-1]++;
				else if (j < 0) distrib[0]++;
				else distrib[j]++; */
				M_tmp += mass;
				mup = ml;
			 };
		 };
	 };
/*
	an = mn - 1;
	printf("# %d  %e  %e  %e  %e  %e\n\n", i, mup, M_tot / N_tot, M_tmp - M_tot, mlim[0], mlim[an]);
	for (i = 0; i < DIST_BIN; i++)
	 {
		mass = mlim[0] * exp((i + 0.5) * log(mlim[an] / mlim[0]) / DIST_BIN);
		printf("%8.4f  %6d\n", mass, distrib[i]);
	 };
*/
	return 0;
};
